@@ -23,6 +23,7 @@ from __future__ import division, print_function, absolute_import
 
 import re
 import ast
+import uuid
 import types
 import hashlib
 import inspect
@@ -32,8 +33,8 @@ from functools import wraps
 from hypothesis.configuration import storage_directory
 from hypothesis.vendor.pretty import pretty
 from hypothesis.internal.compat import ARG_NAME_ATTRIBUTE, hrange, \
-    to_str, qualname, getargspec, to_unicode, isidentifier, str_to_bytes, \
-    update_code_location
+    to_str, qualname, to_unicode, isidentifier, str_to_bytes, \
+    getfullargspec, update_code_location
 
 
 def fully_qualified_name(f):
@@ -45,6 +46,21 @@ def fully_qualified_name(f):
         return qualname(f)
 
 
+def is_mock(obj):
+    """Determine if the given argument is a mock type.
+
+    We want to be able to detect these when dealing with various test
+    args. As they are sneaky and can look like almost anything else,
+    we'll check this by looking for random attributes.  This is more
+    robust than looking for types.
+
+    """
+    for _ in range(10):
+        if not hasattr(obj, str(uuid.uuid4())):
+            return False
+    return True
+
+
 def function_digest(function):
     """Returns a string that is stable across multiple invocations across
     multiple processes and is prone to changing significantly in response to
@@ -68,12 +84,38 @@ def function_digest(function):
     except AttributeError:
         pass
     try:
-        hasher.update(str_to_bytes(repr(getargspec(function))))
+        hasher.update(str_to_bytes(repr(getfullargspec(function))))
     except TypeError:
         pass
     return hasher.digest()
 
 
+def required_args(target, args=(), kwargs=()):
+    """Return a set of names of required args to target that were not supplied
+    in args or kwargs.
+
+    This is used in builds() to determine which arguments to attempt to
+    fill from type hints.  target may be any callable (including classes
+    and bound methods).  args and kwargs should be as they are passed to
+    builds() - that is, a tuple of values and a dict of names: values.
+
+    """
+    try:
+        spec = getfullargspec(
+            target.__init__ if inspect.isclass(target) else target)
+    except TypeError:  # pragma: no cover
+        return None
+    # self appears in the argspec of __init__ and bound methods, but it's an
+    # error to explicitly supply it - so we might skip the first argument.
+    skip_self = int(inspect.isclass(target) or inspect.ismethod(target))
+    # Start with the args that were not supplied and all kwonly arguments,
+    # then remove all positional arguments with default values, and finally
+    # remove kwonly defaults and any supplied keyword arguments
+    return set(spec.args[skip_self + len(args):] + spec.kwonlyargs) \
+        - set(spec.args[len(spec.args) - len(spec.defaults or ()):]) \
+        - set(spec.kwonlydefaults or ()) - set(kwargs)
+
+
 def convert_keyword_arguments(function, args, kwargs):
     """Returns a pair of a tuple and a dictionary which would be equivalent
     passed as positional and keyword args to the function. Unless function has.
@@ -81,11 +123,11 @@ def convert_keyword_arguments(function, args, kwargs):
     **kwargs the dictionary will always be empty.
 
     """
-    argspec = getargspec(function)
+    argspec = getfullargspec(function)
     new_args = []
     kwargs = dict(kwargs)
 
-    defaults = {}
+    defaults = dict(argspec.kwonlydefaults or {})
 
     if argspec.defaults:
         for name, value in zip(
@@ -110,7 +152,7 @@ def convert_keyword_arguments(function, args, kwargs):
                     arg_name
                 ))
 
-    if kwargs and not argspec.keywords:
+    if kwargs and not argspec.varkw:
         if len(kwargs) > 1:
             raise TypeError('%s() got unexpected keyword arguments %s' % (
                 function.__name__, ', '.join(map(repr, kwargs))
@@ -130,11 +172,12 @@ def convert_positional_arguments(function, args, kwargs):
     new_args will only be non-empty if function has a variadic argument.
 
     """
-    argspec = getargspec(function)
-    new_kwargs = dict(kwargs)
-    if not argspec.keywords:
+    argspec = getfullargspec(function)
+    new_kwargs = dict(argspec.kwonlydefaults or {})
+    new_kwargs.update(kwargs)
+    if not argspec.varkw:
         for k in new_kwargs.keys():
-            if k not in argspec.args:
+            if k not in argspec.args and k not in argspec.kwonlyargs:
                 raise TypeError(
                     '%s() got an unexpected keyword argument %r' % (
                         function.__name__, k
@@ -147,6 +190,9 @@ def convert_positional_arguments(function, args, kwargs):
                 raise TypeError('No value provided for argument %s' % (
                     argspec.args[i],
                 ))
+    for kw in argspec.kwonlyargs:
+        if kw not in new_kwargs:
+            raise TypeError('No value provided for argument %s' % kw)
 
     if len(args) > len(argspec.args) and not argspec.varargs:
         raise TypeError(
@@ -201,19 +247,29 @@ def extract_lambda_source(f):
     sins, oh lord
 
     """
-    args = getargspec(f).args
+    argspec = getfullargspec(f)
     arg_strings = []
     # In Python 2 you can have destructuring arguments to functions. This
     # results in an argspec with non-string values. I'm not very interested in
     # handling these properly, but it's important to not crash on them.
     bad_lambda = False
-    for a in args:
+    for a in argspec.args:
         if isinstance(a, (tuple, list)):  # pragma: no cover
             arg_strings.append('(%s)' % (', '.join(a),))
             bad_lambda = True
         else:
             assert isinstance(a, str)
             arg_strings.append(a)
+    if argspec.varargs:
+        arg_strings.append('*' + argspec.varargs)
+    elif argspec.kwonlyargs:
+        arg_strings.append('*')
+    for a in (argspec.kwonlyargs or []):
+        default = (argspec.kwonlydefaults or {}).get(a)
+        if default:
+            arg_strings.append('{}={}'.format(a, default))
+        else:
+            arg_strings.append(a)
 
     if_confused = 'lambda %s: <unknown>' % (', '.join(arg_strings),)
     if bad_lambda:  # pragma: no cover
@@ -246,7 +302,7 @@ def extract_lambda_source(f):
     all_lambdas = extract_all_lambdas(tree)
     aligned_lambdas = [
         l for l in all_lambdas
-        if args_for_lambda_ast(l) == args
+        if args_for_lambda_ast(l) == argspec.args
     ]
     if len(aligned_lambdas) != 1:
         return if_confused
@@ -260,10 +316,9 @@ def extract_lambda_source(f):
             parsed = ast.parse(source[:i])
             assert len(parsed.body) == 1
             assert parsed.body
-            if not isinstance(parsed.body[0].value, ast.Lambda):
-                continue
-            source = source[:i]
-            break
+            if isinstance(parsed.body[0].value, ast.Lambda):
+                source = source[:i]
+                break
         except SyntaxError:
             pass
     lines = source.split('\n')
@@ -304,7 +359,7 @@ def arg_string(f, args, kwargs, reorder=True):
     if reorder:
         args, kwargs = convert_positional_arguments(f, args, kwargs)
 
-    argspec = getargspec(f)
+    argspec = getfullargspec(f)
 
     bits = []
 
@@ -315,10 +370,7 @@ def arg_string(f, args, kwargs, reorder=True):
         for a in sorted(kwargs):
             bits.append('%s=%s' % (a, nicerepr(kwargs[a])))
 
-    return ', '.join(
-        [nicerepr(x) for x in args] +
-        bits
-    )
+    return ', '.join([nicerepr(x) for x in args] + bits)
 
 
 def unbind_method(f):
@@ -373,8 +425,8 @@ def define_function_signature(name, docstring, argspec):
         check_valid_identifier(a)
     if argspec.varargs is not None:
         check_valid_identifier(argspec.varargs)
-    if argspec.keywords is not None:
-        check_valid_identifier(argspec.keywords)
+    if argspec.varkw is not None:
+        check_valid_identifier(argspec.varkw)
     n_defaults = len(argspec.defaults or ())
     if n_defaults:
         parts = []
@@ -384,11 +436,14 @@ def define_function_signature(name, docstring, argspec):
             parts.append('%s=not_set' % (a,))
     else:
         parts = list(argspec.args)
-    used_names = list(argspec.args)
+    used_names = list(argspec.args) + list(argspec.kwonlyargs)
     used_names.append(name)
 
+    for a in argspec.kwonlyargs:
+        check_valid_identifier(a)
+
     def accept(f):
-        fargspec = getargspec(f)
+        fargspec = getfullargspec(f)
         must_pass_as_kwargs = []
         invocation_parts = []
         for a in argspec.args:
@@ -400,13 +455,21 @@ def define_function_signature(name, docstring, argspec):
             used_names.append(argspec.varargs)
             parts.append('*' + argspec.varargs)
             invocation_parts.append('*' + argspec.varargs)
+        elif argspec.kwonlyargs:
+            parts.append('*')
         for k in must_pass_as_kwargs:
             invocation_parts.append('%(k)s=%(k)s' % {'k': k})
 
-        if argspec.keywords:
-            used_names.append(argspec.keywords)
-            parts.append('**' + argspec.keywords)
-            invocation_parts.append('**' + argspec.keywords)
+        for k in argspec.kwonlyargs:
+            invocation_parts.append('%(k)s=%(k)s' % {'k': k})
+            if k in (argspec.kwonlydefaults or []):
+                parts.append('%(k)s=not_set' % {'k': k})
+            else:
+                parts.append(k)
+        if argspec.varkw:
+            used_names.append(argspec.varkw)
+            parts.append('**' + argspec.varkw)
+            invocation_parts.append('**' + argspec.varkw)
 
         candidate_names = ['f'] + [
             'f_%d' % (i,) for i in hrange(1, len(used_names) + 2)
@@ -427,6 +490,10 @@ def define_function_signature(name, docstring, argspec):
         result = base_accept(f)
         result.__doc__ = docstring
         result.__defaults__ = argspec.defaults
+        if argspec.kwonlydefaults:
+            result.__kwdefaults__ = argspec.kwonlydefaults
+        if argspec.annotations:
+            result.__annotations__ = argspec.annotations
         return result
     return accept
 
@@ -453,7 +520,6 @@ def impersonate(target):
 
 def proxies(target):
     def accept(proxy):
-        return impersonate(target)(wraps(target)(
-            define_function_signature(
-                target.__name__, target.__doc__, getargspec(target))(proxy)))
+        return impersonate(target)(wraps(target)(define_function_signature(
+            target.__name__, target.__doc__, getfullargspec(target))(proxy)))
     return accept
